module envision.blitter;

private import envision.canvas;

struct Opaque
{
    // src is the pixel to be plotted, dest is the pixel already on the canvas.
    // The return value is the two colours blended.
    static Colour blend(Colour src, Colour dest)
    {
        return src;
    }
}

struct Matte
{
    static Colour blend(Colour src, Colour dest)
    {
        if (src.a != 0)
            return src;
        else
            return dest;
    }
}

struct AlphaBlend
{
    static Colour blend(Colour src, inout Colour dest)
    {
        if (src.a == 0)
            return dest;
        else if (dest.a == 0)
            return src;
        else
        {
            ubyte a = src.a;
            ubyte ia = 255 - a;
            dest.r = (src.r * a + dest.r * ia) / 65535;
            dest.g = (src.g * a + dest.g * ia) / 65535;
            dest.b = (src.b * a + dest.b * ia) / 65535;
            return dest;
        }
    }
}

    // too many arguments -_-
private void doClipping(int xmin, int ymin,
                int xmax, int ymax,
                inout int x, inout int y,
                out int xstart, out int ystart,
                inout int xlen, inout int ylen)
{
    xstart = ystart = 0;

    if (x < xmin)
    {
        xlen   -= xmin - x;
        xstart += xmin - x;
        x = xmin;
    }

    if (y < ymin)
    {
        ylen   -= ymin - y;
        ystart += ymin - y;
        y = ymin;
    }

    if (x + xlen > xmax)
        xlen = xmax - x;
    if (y + ylen > ymax)
        ylen = ymax - y;
}


template blit(Blender)
{
    void blit(Canvas src, Canvas dest, int x, int y)
    {
        int xlen = src.width();
        int ylen = src.height();

        int xstart = 0;
        int ystart = 0;

        doClipping(
            0, 0,                           // xmin, ymin
            dest.width(), dest.height(),    // xmax, ymax
            x, y,                           // x, y
            xstart, ystart,                 // xstart, ystart
            xlen, ylen);                    // xlen, ylen

        if (xlen < 1 || ylen < 1)   return; // all clipped

        while (ylen > 0)
        {
            Colour* psrc  = src.getPointer(xstart, ystart);
            Colour* pdest = dest.getPointer(x, y);

            int curx = xlen;
            while (curx--)
            {
                *pdest = Blender.blend(*psrc, *pdest);
                psrc++;
                pdest++;
            }

            y++;
            ylen--;
            ystart++;
        }
    }
}

// TODO: scale blit, rotate blit, et cetera
// also primitives. (lines, dots, rectangles)

alias blit!(Opaque) opaqueBlit;
alias blit!(Matte)  matteBlit;
alias blit!(AlphaBlend) alphaBlit;
